/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

#include "backtalk.h"
#include "udb.h"
#include "file.h"
#include "tag.h"
#include "str.h"
#include "vstr.h"
#include "readfile.h"
#include "last_log.h"
#include "partdir.h"
#include "btpasswd.h"
#include "group.h"
#include "partfile.h"
#include "dumputil.h"

#define wst(n,v) wrtstrtag(stdout,n,v,FALSE)
#define wit(n,v) wrtinttag(stdout,n,v,FALSE)

#include <string.h>

#ifdef HAVE_DIRENT_H
#include <dirent.h>
typedef struct dirent direntry;
#else
typedef struct direct direntry;
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#endif


/* XMLQUOTE - Expand " to &quot; and so forth.  We'd use sysformat() but
 * that is so riddled with references to the dictionary and the stack that
 * it doesn't embed well in stand alone programs.  So we write a little
 * clean version.  Returns the expanded string in dynamic memory, which
 * the caller must free.
 */

char *xmlquote(char *s)
{
    VSTRING out;

    vs_new(&out, 0);

    for (; *s != '\0'; s++)
	switch (*s)
	{
	    case '<':  vs_cat_n(&out,"&lt;",  4); break;
	    case '>':  vs_cat_n(&out,"&gt;",  4); break;
	    case '&':  vs_cat_n(&out,"&amp;", 5); break;
	    case '"':  vs_cat_n(&out,"&quot;",6); break;
	    case '\'': vs_cat_n(&out,"&apos;",6); break;
	    default:   *vs_inc(&out)= *s;         break;
	}
    return vs_return(&out);
}


/* WRITEHASHDATA - Write out tag lines for everything in the given hash */

void writehashdata(HashTable *hash)
{
    HashEntry *h;
    HashSearch hs;

    for (h= FirstHashEntry(hash, &hs); h != NULL; h= NextHashEntry(&hs))
	wrttag(stdout, h->key, &(h->t));
}


/* WRTTAG - Duplicate of the version in file.c so we don't have to link in
 * file.c
 */

void wrttag(FILE *fp, char *name, Token *t)  /* DUPLICATE */
{
    if (type(*t) == TK_STRING)
	wrtstrtag(fp,name,sval(*t),FALSE);
    else if (type(*t) == TK_INTEGER)
	wrtinttag(fp,name,ival(*t),FALSE);
    else
	die("%s is not a string or integer variable", name);
}


/* WRITEFILE - Given a full path and an identifier for a file, write it
 * out surrounded by <file> tags.  If the file is not readable, write nothing.
 * Quote file and make sure it ends with a newline.
 */

void writefile(char *fname, char *what)
{
    int i;
    char *content, *quoted;

    content= read_file(fname,FALSE);
    if (content == NULL || *content == '\0')
	return;
    printf("<file name=\"%s\">\n", what);
    quoted= xmlquote(content);
    free(content);
    fputs(quoted, stdout);
    i= strlen(quoted);
    if (quoted[i-1] != '\n') putchar('\n');
    free(quoted);
    printf("</file>\n", what);
}


/* WRITEPF - Write out a description of the participation file for the given
 * user with the given name.  The path of the user's participation file
 * directory is given.
 */

void writepf(struct btpasswd *pw, char *pfpath, char *pfname)
{
    struct plist *pl;
    struct pfc pf;
    struct pfe *item;
    char *f;

    /* See if we have a conference for this file first */
    if ((pl= findpart(pfname)) == NULL)
	fprintf(stderr,"Skipping participation file %s/%s - "
		"not for any known conference", pfpath, pfname);

    /* Make a participation structure for this conference */
    pf.snooping= TRUE;
    strcpy(pf.part_name,pfpath);
    strcat(pf.part_name,"/");
    strcat(pf.part_name,pfname);
    pf.user= pw->name;
    pf.conf= pl->cf;
    pf.confdir= pl->dr;

    /* Load it in */
    if (loadpart(&pf)) return;

    /* Print it out */
    printf("<participation conf=\"%s\">\n", f= xmlquote(pl->cf));
    free(f);

    if (pf.alias != NULL || pf.conf_etc != NULL)
	printf("<data>\n");
    if (pf.alias != NULL)
	wst("alias",pf.alias);
    if (pf.conf_etc != NULL)
	writehashdata(pf.conf_etc);
    if (pf.alias != NULL || pf.conf_etc != NULL)
	printf("</data>\n");

    for (item= pf.pflist; item != NULL; item= item->next)
    {
	printf("<item number=\"%d\">\n", item->item_no);
	wit("lastresp", item->resp_no);
	wit("lastdate",item->date);
	if (item->forgotten) wit("forgotten",1);
	if (item->favorite) wit("favorite",1);
	if (item->my_title != NULL) wst("title",item->my_title);
	if (item->etc != NULL) writehashdata(item->etc);
	printf("</item>\n");
    }

    printf("</participation>\n");

    /* Deallocate stuff */
    freepart(&pf);
}

void dump_user(struct btpasswd *pw, int dump_pass, int save_part)
{
    char *f, *g;
    rtag *rt;
    char rc[BFSZ];
#ifdef PART_DIR
    char pf[BFSZ];
#else
    char *pf;
#endif
    int rclen,pflen;

    if (pw == NULL) return;

    printf("<user login=\"%s\">\n",
	f= xmlquote(pw->name), pw->uid, pw->gid);
    free(f);
    printf("<data>\n");
    if (pw->gecos != NULL && *pw->gecos != '\0')
	wst("usrname",pw->gecos);
    wit("usrid",pw->uid);
    wst("grpname",groupname(pw->gid));
    g= get_user_groups(pw->name, pw->gid);
    if (g != NULL)
    {
	wst("grps",g);
	free(g);
    }
    if (dump_pass)
    {
	int status;
	char *p= pw->passwd;

	/* If we don't already have the password, try harder */
	if (p == NULL || *p == '\0')
	    p= getpassword(pw->name, FALSE);

	if (p != NULL && *p != '\0')
	{

	    if (*p == '?')
	    {
		status= 1; p++;
	    }
	    else if (*p == '*')
	    {
		status= 2; p++;
	    }
	    else
		status= 0;

	    wst("encpasswd",p);
	    if (status != 0) wit("usrstat", status);
	}
    }
    wst("dirpath",pw->dir);

    wit("privacy",pw->priv);
    if (pw->email && pw->email[0]) wst("email",pw->email);
    if (pw->valcode && pw->valcode[0]) wst("evalid",pw->valcode);
    wit("eprivacy",pw->epriv);

    /* Get last log time */
#ifdef LASTLOG_FILE
    wit("laston",get_lastlog(pw->uid));
#endif

    /* Construct directory path for .cfonce and .cfrc */
    rc_dir(rc, pw);
    rclen= strlen(rc);
    /* Construct directory path for participation files and .backtalk */
#ifdef PART_DIR
    homepartpath(pf, pw->dir, "");
    pflen= strlen(pf);
#else
    pf= rc;
    pflen= rclen;
#endif

    /* Load .backtalk file */
    strcpy(pf+pflen, ".backtalk");
    if ((rt= open_tag(pf,TRUE)) != NULL)
    {
	int code;
	char *var, *val;

	while ((code= get_tag(rt, &var, &val)) > 0)
	{
	    switch (code)
	    {
		case 1: wst(var,val); break;
		case 2: wit(var,atoi(val)); break;
		case 3: wrtstrtag(stdout,var,val,TRUE);
		case 4: wrtinttag(stdout,var,atoi(val),TRUE);
	    }
	}
    }
    printf("</data>\n");
    
    /* Load .cflist file */
    strcpy(pf+pflen, ".cflist");
    writefile(pf, ".cflist");

    /* Load .cfonce file */
    strcpy(rc+rclen, ".cfonce");
    writefile(rc, ".cfonce");

    /* Load .cfrc file */
    strcpy(rc+rclen, ".cfrc");
    writefile(rc, ".cfrc");

    if (save_part)
    {
	DIR *dp;
	direntry *file;

	pf[pflen]= '\0';
	if ((dp= opendir(pf)) != NULL)
	{
	    while ((file= readdir(dp)) != NULL)
	    {
		int i;
		char *f= file->d_name;

		if (f[0] != '.') continue;
		i= strlen(f);
		if (i >= 5 && f[i-3] == '.' &&
			      f[i-2] == 'c' &&
			      f[i-1] == 'f')
		    writepf(pw, pf, f);
	    }
	}
    }

    printf("</user>\n");
}


int main(int argc, char **argv)
{
    struct btpasswd *pw;
    int dump_pass= TRUE;
    int save_part= TRUE;
    struct usl {
	char *u;
	struct usl *next;
    } *u, *last= NULL, *userlist= NULL;

    int i,j;

    for (i= 1; i < argc; i++)
    {
	if (argv[i][0] == '-')
	    for (j= 1; argv[i][j] != '\0'; j++)
		switch (argv[i][j])
		{
		    case 'p':  save_part= FALSE; break;
		    case 'P':  dump_pass= FALSE; break;
		    default: goto usage;
		}
	else
	{
	    /* Insert user into user name list */
	    u= (struct usl *)malloc(sizeof(struct usl));
	    u->u= argv[i];
	    u->next= NULL;
	    if (last == NULL)
		userlist= u;
	    else
		last->next= u;
	    last= u;
	}
    }

    if (save_part)
    {
	/* Build a list of conference names/directories that go with
	 * each participation file name.
	 */
	mkpartlist();
    }

    /* Initialize the group list structure */
    for_groups(add_group);

    printf("<userdata date=\"%ld\" version=\"%d.%d.%d\">\n", time(0L),
	    VERSION_A, VERSION_B, VERSION_C);

    /* Looop through the users */
    if (userlist == NULL)

	/* All users */
	for (pw= walkdb(0); pw != NULL; pw= walkdb(1))
	    dump_user(pw, dump_pass, save_part);

    else

	/* Just listed users */
	for (u= userlist; u != NULL; u= u->next)
	{
	    struct btpasswd *pw= getdbnam(u->u);
	    if (pw == NULL) fprintf(stderr, "user %s not found\n", u->u);
	    dump_user(pw, dump_pass, save_part);
	}

    printf("</userdata>\n");

    exit(0);
usage:
    fprintf(stderr,"usage: %s [-pP] [<user>...]\n", argv[0]);
    exit(1);
}


/* Some dummy routines that never get called, but need to exist to let
 * things link.  Most of these are called by newpart() which will never
 * be called because we always open participation files in snoop mode.
 */

void seek_item(int x) {}   /* DUMMY */
int next_item(int *a,int *b,int *c,time_t *d,time_t *e) {return 1;}  /*DUMMY*/
int in_sel(int i, char *s) {return 0;} /* DUMMY */
char *user_fname() {return "";} /* DUMMY */
char *get_noise() {return NULL;} /* DUMMY */
